#!/usr/bin/python

# Copyright (C) 2014 - 2018 by Geoffrey Prewett.
# Permission granted for personal and commercial use, except for selling and
# distributing this script.  You may distribute and/or sell anything created
# with this software, just not this script itself.

import copy
import errno
import keyword # for parsing
import glob
import os
import os.path
import re
import sys
import traceback

import urllib

ERR = sys.stderr

kTypeString = "_string"
kTypeArray = "_array"
kTypeThis = "_this"

kThisKey = "\\template"
kCurrentIndexKey = "\\currentIndex"
kCurrentIndexIsEvenKey = "\\currentIndexIsEven"
kCurrentIndexIsOddKey = "\\currentIndexIsOdd"
kIndexKey = "\\index"
kIndexEntryKey = "\\indexEntry"
kIndexEntryURLKey = "\\indexEntryURL"

kEvalPrefix = "\\__evalAtLine"

kIndexFilename = "index.html"

kBeginTemplate = "\\begin "
kEndTemplate = "\\end "
kBeginCodeRE = re.compile(r"\\beginCode")
kEndCodeRE = re.compile(r"\\endCode")
kEndCode = "\\endCode"
kOneLineTemplate = "\\define"
kOneLineTemplateRE = re.compile('^\\' + kOneLineTemplate + '\\s+([a-zA-Z_\\-0-9]+)\\s+(.*)\\n$')
kDataLineRE = re.compile('^\\s*(\\\[a-zA-Z_\\-0-9]+)\\s+(.*)$')
kDataOpenRE = re.compile(r'^\s*(\\[a-zA-Z0-9_\\-]+)\s+{\s*$')
kDataCloseRE = re.compile('^\\s*}\\s*$')
kArrayCloseRE = re.compile('^\\s*\\]\\s*$')
#kDataFindVarRE = re.compile('(\\[a-zA-Z_\\-]+)')
kKeyRE = re.compile('(\\\\[a-zA-Z_\\-0-9]+)')
# Don't need to match to \s, but if we do, then match.end() + 1 is the start
# of the rest of the line
kCmdRE = re.compile(r'^\s*(\\[A-Za-z0-9\\-_]+)\s*')
kEndForRE = re.compile(r'\s*\\endfor\s*$')
kIncludeRE = re.compile('^\\\\include\\s+([^\\n\\r]+)$')
kVariableFuncRE = re.compile(r'(\\{?[a-zA-Z0-9_\-]+)(\.([a-zA-Z0-9_\-]+)(\(([^\)]*)\))?)*}?')

# Can't match HTML comments: <!-- and -->
kReplaceTwoDashesRE = re.compile('(?!<\\!)--(?!>)')

#kReplaceLeftDoubleQuoteRE = re.compile('(?:>|.*\\s)(")[A-Za-z]')

kKeywords = frozenset(['\\begin', '\\end', '\\option', '\\if', '\\ifnot', '\\ifexists', '\\else', '\\endif', '\\for', '\\endfor', '\\evalfile', '\\include', '\\eval', '\\beginCode', '\\endCode'])

gIsVerbose = False
gFileCache = {}
gIncludeDirs = []

class Context:
    def __init__(self, data):
        self.data = data

    def added(self, key, value):
        c = Context(dict(self.data))
        c.data[key] = value
        return c

def error(context, error):
    errStr = ""
    filePath = ""
    if "cwd" in context.data:
        filePath = calcRelativePath(context.data["cwd"],
                                    context.data["filename"])
    if filePath == "":
        filePath = context.data["filename"]

    templateLineNo = 0
    if "templateLineNo" in context.data:
        templateLineNo = context.data["templateLineNo"]

    errStr += filePath
    if "lineNo" in context.data:
        lineNo = context.data["lineNo"]
        if lineNo > 0:  # 0 is default from *Value constructors
            if "template" in context.data:
                lineNo += templateLineNo
            errStr += ":" + str(lineNo)

    if "template" in context.data:
        errStr += " (" + context.data["template"]
        if "templateLineNo" in context.data:
            if templateLineNo > 0:
                # Add one because the \begin line isn't counted internally
                # (the string starts on the next line), but users will
                # will probably count it as the first line
                errStr += ":" + str(templateLineNo + 1)
        errStr += ")"

    errStr += ": " + error
    print >>ERR, errStr

def main():
    global gIncludeDirs, gIsVerbose

    isSpecificFile = False
    isRecursive = False
    cwd = os.getcwd()
    dataFiles = []
    templateFiles = []

    idx = 1  # skip argv[0]!
    while idx < len(sys.argv):
        if sys.argv[idx] == "-I":
            idx += 1
            d = sys.argv[idx]
            if d[0] != '/':
                d = cwd + '/' + d
            templateFiles += [d + "/" + f for f in os.listdir(d) if f.endswith(".tt")]
            gIncludeDirs.append(d)
        elif sys.argv[idx] == "-r":
            isRecursive = True
        elif sys.argv[idx] == "-v":
            gIsVerbose = True
        else:
            dataFiles.append(sys.argv[idx])
        idx += 1
#    if len(dataFiles) == 0:
#        dataFiles = [f for f in os.listdir(cwd) if f.endswith(".t")]
#    else:
#        isSpecificFile = True  # Can't just check dataFiles, might be a dir with
#                               # only one .t file in it
    if len(dataFiles) == 0:
        dataFiles = None
    else:
        isSpecificFile = True

    templates = loadTemplateFiles({}, templateFiles, cwd)
    processDirectory(templates, cwd, dataFiles, isRecursive=isRecursive)

def processDirectory(templates, path, dataFiles, isRecursive=False):
    templates = copy.copy(templates)
    templateFiles = [path + "/" + f for f in os.listdir(path) if f.endswith(".tt")]

    templates = loadTemplateFiles(templates, templateFiles, path)

    indexData = None
    if dataFiles is None and kIndexKey in templates and kIndexEntryKey in templates:
        indexData = []
        if templates[kIndexKey].indexSubdirs() == True:
            dataFiles = [y for x in os.walk(path) for y in glob.glob(os.path.join(x[0], '*.t'))]

    if dataFiles is None:
        dataFiles = [path + "/" + f for f in os.listdir(path) if f.endswith(".t")]
    for f in dataFiles:
        context = Context({ "cwd": os.path.dirname(f), "filename": f })
        text, data = processDataFile(templates, f, context)
        if text is not None:
            outpath = calcHTMLFilename(f)
#            print >>sys.stderr, "[debug]", f, "->", outpath
            writeFile(outpath, text, f)
        if data is not None and indexData is not None:
            data.data[kIndexEntryURLKey] = StringValue(kTypeString, calcHTMLFilename(f), f)
            indexData.append(StructValue(kIndexEntryKey, data.data, f))
            
    if indexData is not None:
        indexFilename = path + "/" + kIndexFilename
        if gIsVerbose:
            print "Writing index to", indexFilename
        data = StructValue(kIndexKey, { kIndexKey: ArrayValue(indexData, f) }, f)
        context = Context({ "cwd": path, "filename": indexFilename,
                            "template": kIndexKey })
        index = instantiateTemplate(templates, templates[kIndexKey], data,
                                    context)
        writeFile(indexFilename, index, context)

    if isRecursive:
        # An index template is only good for the current directory, so remove
        # it if we have one.
        if kIndexKey in templates:
            templates.pop(kIndexKey)
        if kIndexEntryKey in templates:
            templates.pop(kIndexEntryKey)

        for d in os.listdir(path):
            dPath = path + "/" + d
            if d != "." and d != ".." and os.path.isdir(dPath):
                processDirectory(templates, dPath, None, isRecursive)

def loadTemplateFiles(templates, templateFiles, cwd):
    for f in templateFiles:
        # skip hidden files (such as Emacs' temporaries)
        if f[0] == '.' and not f.startswith('./') and not f.startswith('../'):
            continue
        context = Context({ "cwd": cwd, "filename": f })
        templates = loadTemplateFile(templates, f, context)
    return templates

def loadTemplateFile(templates, path, context):
    # TODO: this could return a key, in which case this is not a pure template
    #       file and should be handled accordingly
    k, v = readDataFile(templates, path, context)

    return templates

def isStartOfTemplate(line):
    if line.startswith(kBeginTemplate) or line.startswith(kOneLineTemplate+ " "):
        return True
    return False

def parseTemplate(filename, textIt, templates, line=None):
    name = ""
    value = ""
    nErrors = 0
    startLineNo = textIt.lineNo()
    if line is None:
        line = textIt.nextLine()
    while not textIt.isDone():
        if line.startswith(kBeginTemplate):
            words = line.split()
            if len(words) < 2:
                error(Context({"filename": filename, "lineNo": textIt.lineNo()}),
                      "syntax error, expecting:  \\begin name \\n")
                name = path + ":" + str(lineNo)
                nErrors += 1
            if len(words) >= 2:
                name = words[1]
            if name in templates:
                error(Context({"filename": filename, "lineNo": textIt.lineNo()}),
                      "a template named '" + name +"' already exists;  ignoring")
                nErrors += 1
        elif line.startswith(kEndTemplate):
            return (name, Template(filename, name, startLineNo, value), nErrors)
        # Note:  lines starting with \ are legal:
        #        <div style="...">
        #        \key-to-substitute
        #        </div>
#        elif line.startswith('\\'):
#            error("syntax error: " + line.split()[0] + "is not a command")
        else:
            onelineResults = kOneLineTemplateRE.search(line)
            if onelineResults is not None:
                return (onelineResults.group(1), onelineResults.group(2),
                        nErrors)
            else:
                value += line  # keep the \n (so result is readable)
        line = textIt.nextLine()
    return (name, Template(filename, name, startLineNo, value), nErrors)

def processDataFile(templates, path, context):
    global gFileCache

    if path in gFileCache:
        if isinstance(gFileCache[path], basestring):
            return gFileCache[path]

    key, data = readDataFile(templates, path, context)
    if key is None or data is None:
        return (None, None)
    context = context.added("template", key).added("lineNo", data.lineNo)

    isArray = isinstance(data, ArrayValue)

    if key in templates:
        templ = templates[key]
    if key not in templates:
        if isArray:
            templ = Template(path, key, data.filename, data.lineNo)
        else:
            error(context, "No template '" + str(key) + "'")
            return ("", StructValue("error", {}, context))

    # Do all the templating
    if isArray:
        text = evalArray(templates, templ, data, context, False)
    else:
        text = instantiateTemplate(templates, templ, data, context)

    # If there is anything left, must be a template (for example, a \define)
    keyResult = kKeyRE.search(text)
    while keyResult is not None:
        k = keyResult.group(1)
        if k in templates:
            text = text.replace(k, templates[k])
        else:
            text = text.replace(k, k[1:].upper())
        keyResult = kKeyRE.search(text)

    text = beautifyHTML(text)

    if path not in gFileCache:
        gFileCache[path] = True
    else:
        gFileCache[path] = (text, data)

    return (text, data)

def readDataFile(templates, path, context):
    filename = path
    data = {}
    key = None
    nErrors = 0
    useNewFileTemplates = {}
    indexSubdirs = False
    try:
        IN = open(path, "r")
        lines = IN.readlines()
        IN.close()
    except IOError, e:
        error(context, e.message)
        return (None, None)

    textIt = TextIterator(lines)
    while not textIt.isDone():
        line = textIt.nextLine()
        if line.startswith("\\include "):
            inclFile = kIncludeRE.search(line).group(1)
            line = textIt.includeFile(inclFile, context.data["cwd"], context)

        if isStartOfTemplate(line):
            name, template, nerrs = parseTemplate(filename, textIt, templates, line=line)
            if name in useNewFileTemplates:
                template.setNewFilePrefix(useNewFileTemplates[name])
            if indexSubdirs:
                template.setIndexSubdirs(True)
            if name not in templates:
                templates["\\" + name] = template
            else:
                error(context, "Template '" + name + "' has already been defined in '" + template["\\" + name].fullPath() + "'")
            nErrors += nerrs
        elif line.isspace():
            pass
        elif line.startswith("\\option "):
            args = line.split()
            if len(args) < 2:
                error(context, "Bad syntax for \\option")
                continue
            if args[1] == "ignore":
                return (None, None)
            elif args[1] == "useNewFileForTemplateArray":
                if len(args) > 4:
                    error(context, "Bad syntax, should be:  \\option useNewFileForTemplateArray templateName [prefix]")
                if len(args) == 3:
                    useNewFileTemplates[args[2]] = args[2]
                elif len(args) == 4:
                    useNewFileTemplates[args[2]] = args[3]
            elif args[1] == "indexSubdirectories":
                indexSubdirs = True
        else:
            # Hack:  parseData returns StructValue, but expects dict,
            #        so more than one data object is in a file, we need to
            #        extract the dictionary.
            if isinstance(data, StructValue):
                data = data.data
            key, data, nErrors = parseData(context, textIt, data, line=line)

#    if key not in templates:   # use str(key) in case of None or something...
#        error(context, "No template available for '" + str(key) + "'")
#        return (None, None)

    if nErrors > 0:
        return (None, None)

    return key, data

def beautifyHTML(html):
    # Make some typographic fixes
    # Replace -- needs to consider <!--
    idx = html.find('</head>')
    if idx < 0:
        idx = 0;
    else:
        idx += 8
    while idx >= 0:
        nextEntityIdx = html.find('<', idx)
        if nextEntityIdx != idx:   # Also includes -1
            endIdx = nextEntityIdx
            if nextEntityIdx < 0:
                endIdx = len(html)
            subtext = beautifyRawText(html[idx:endIdx])
            if nextEntityIdx > 0:
                html = html[:idx] + subtext + html[endIdx:]
            else:
                html = html[:idx] + subtext
                break
            idx += len(subtext)
        idx = html.find('>', idx)
        if idx > 0:
            idx += 1

    return html

def beautifyRawText(text):
    # Substitute -- with a real dash
    text = kReplaceTwoDashesRE.sub('&mdash;', text)
    # Substitute '. &nbsp;' and '.&nbsp; '
    # Note:  sometimes \n is used as the space
    # Note:  do this before ' and " get substituted!
    text = re.sub(r'([\.\?\;\:\,\!\"\'])(\s&nbsp;|&nbsp;\s)', r'\1 ', text)
    # Substitute double-quotes with open/close quotes
    text = re.sub(r'(^|\s|[\(\[])(")([A-Za-z0-9\(\[\.])', r'\1&ldquo;\3', text)
    text = re.sub(r'([A-Za-z\.\,\!\?\)\]])(")(\s|(?:&nbsp;)|[\)\]\,\.\:\;\?\!]|$)', r'\1&rdquo;\3', text)
    # Substitute single-quotes with open/close single quotes
    # Note:  '60 is not an open quote
    text = re.sub(r'(^|\s)(\')([A-Za-z\(\[\.])', r'\1&lsquo;\3', text)
    # Note:  if quote is sandwiched by letters, is an apostrophe...
    text = re.sub(r'([A-Za-z\.\,\!\?\)\]])(\')(\s|(?:&nbsp;)|$|[a-z])', r'\1&rsquo;\3', text)
    # Handle quotes within a quote
    text = re.sub(r'(^|[^A-Za-z])"\'', r'\1&ldquo;&lsquo;', text)  # "'
    text = re.sub(r'\'"($|[^A-Za-z])', r'&rsquo;&rdquo;\1', text)  # '"

    return text

def instantiateTemplate(templates, t, keyValues, context, globalVals=None):
    if globalVals is None:
        thisVal = ThisValue(context.data["filename"])
        globalVals = StructValue("", { kThisKey: thisVal }, context.data["filename"])
    cwd = ""
    if "cwd" in context.data:
        cwd = context.data["cwd"]
    context = Context({ "cwd": cwd,
                        "filename": t.path(),
                        "lineNo": t.lineNo(),
                        "template": t.name() })
    return instantiateText(TextIterator(t.content()), t.directory(), templates,
                           keyValues, globalVals, context)

def instantiateText(textIt, pathPrefix, templates, keyValues, globalVals,
                    context, justCommands=False):
    instantiatedLines = []
    while not textIt.isDone():
        line = textIt.nextLine()
        lineNo = textIt.lineNo()
        if "templateLineNo" in context.data:
            lineNo += context.data["templateLineNo"]
        match = kCmdRE.search(line)
        if match is None:  # normal line
            instantiatedLines.append( (line, context.added("templateLineNo",
                                                           lineNo)) )
            continue

        cmd = match.group(1)
        if cmd == "\\include":
            match = kIncludeRE.search(line).group(1)
            if match[0] == '\\':
                if match in keyValues.data:
                    v = keyValues.data[match]
                    includedLines = v.contentAsTextLines()
                    for inclLine in includedLines:
                        instantiatedLines.append( (inclLine, context.added("lineNo", lineNo)) )
                else:
                    error(context.added("lineNo", textIt.lineNo()), "Error including key '" + match + "': key does not exist")
            else:
                inclFile = searchForFile(match.group(1), pathPrefix)
                try:
                    inclContext = Context({ "cwd": "", "filename": inclFile })
                    inclLineNo = 1
                    IN = open(inclFile)
                    for inclLine in IN:
                        instantiatedLines.append( (inclLine, context.added("lineNo", inclLineNo)) )
                        inclLineNo += 1
                    IN.close()
                except IOError, e:
                    error(context.added("lineNo", textIt.lineNo()), "Error including file '" + inclFile + "': " + e.message)
            continue
        elif cmd == "\\evalfile":
            filename = searchForFile(line[match.end():], pathPrefix)
            text, _ = processDataFile(templates, filename, context)
            if text is not None:
                evalContext = Context({ "cwd": "", "filename": filename })
                instantiatedLines.append( (text, evalContext) )
            continue
        elif cmd == "\\for":
            args = line.split()
            if len(args) != 4:
                error(context, "Syntax error;  should be \\for iterationVar in \\contentVar")
                iterationVar = None
                contentVar = None
            else:
                iterationVar = args[1]
                contentVar = args[3]
            forLines = readForLines(textIt, line, cmd, templates,
                                    [keyValues, globalVals], context)
            if iterationVar is None:
                continue
            # TODO fix line numbers in errors in expandVariables()
            forContext = context.added("templateLineNo", lineNo)
            content, _, err = evalVarString(contentVar, [keyValues, globalVals],
                                            context)
            if err:
                error(forContext, contentVar + " does not exist")
                instantiatedLines.append( ("FOR " + contentVar.upper(),
                                           forContext) )
                continue
            if not isinstance(content, ArrayValue):
                error(forContext, contentVar + " is not an array")
                instantiatedLines.append( ("FOR " + contentVar.upper(),
                                           forContext) )
                continue
            iterVarName = "\\" + iterationVar
            if iterationVar[0] == '\\':
                error(forContext, "Iteration variable should not be prefixed with a '\\'")
                iterVarName = iterationVar
            forThisVal = ThisValue(context)
            forThisVal.setLoopCount(len(content.data))
            forGlobalVals = StructValue("", { kThisKey: forThisVal }, context)
            forText = "\n".join(forLines)
            forText = expandVariables(forText, templates, keyValues, None, context, ignoreErrors=True)
            forLineNo = lineNo
            for forIdx in xrange(0, len(content.data)):
                v = content.data[forIdx]
                forKeyValues = StructValue(iterVarName, { iterVarName: v }, context)
                newText = instantiateText(TextIterator(forText), pathPrefix, templates, forKeyValues, forGlobalVals, context.added("templateLineNo", forLineNo))
                if newText != "":
                    instantiatedLines.append( (newText, context.added("templateLineNo", forLineNo)) )
                forThisVal.incrementId()
                forLineNo += 1
            continue
        elif cmd == "\\if" or cmd == "\\ifnot" or cmd == "\\ifexists":
            ifLines = readIfLines(textIt, line, cmd, templates,
                                  [keyValues, globalVals], context)
            ifText = instantiateText(TextIterator(ifLines), pathPrefix, templates, keyValues, globalVals, context, justCommands=True)
            if ifText != "":
                # Insert the lines individually, in case we if'd in some things
                # into a data object, as parseArray assumes one line per item
                for ln in ifText.splitlines():
                    instantiatedLines.append( (ln, context.added("templateLineNo", lineNo)) )
            continue

        # In case isn't a real command (probably variable substitution),
        # treat as normal line.
        instantiatedLines.append( (line, context.added("templateLineNo", lineNo)) )

    # Now go through and find the template instantiations
    textLines = [line[0] for line in instantiatedLines]
    textIt = TextIterator(textLines)
    expandedText = ""
    while not textIt.isDone():
        line = textIt.nextLine()
        ctx = instantiatedLines[textIt.lineNo() - 1][1]
        startLineNo = 0
        if "templateLineNo" in ctx.data:
            startLineNo = ctx.data["templateLineNo"]
        dataMatch = kDataOpenRE.search(line)
        if dataMatch is not None:
            key, val, _, = parseData(context, textIt, {}, line, { "templates": templates, "keyValues": keyValues, "globalVals": globalVals })

            text = instantiateText(TextIterator(key), pathPrefix, templates, val, globalVals, context.added("templateLineNo", startLineNo))
        else:
            text = line

        # It would probably be more efficient to do something like
        # '\n'.join(instantiatedLines) and run expandVariables() on that,
        # but then we have no chance of getting decent line information for
        # errors.
        if not justCommands:
            text = expandVariables(text, templates, keyValues, globalVals, context.added("templateLineNo", startLineNo))

        if len(expandedText) > 0:
            expandedText += "\n"
        expandedText += text

    return expandedText

def readIfLines(textIt, line, cmd, templates, keyValuesArray, context):
    argStartIdx = len(cmd) + 1
    while argStartIdx < len(line) and line[argStartIdx].isspace():
        argStartIdx += 1
    args = [ line[argStartIdx:] ]
    if cmd == "\\ifexists":
        if "\\" + args[0] in keyValuesArray[0].data:
            useThisCase = True
        else:
            useThisCase = False
    elif cmd == "\\if" or cmd == "\\ifnot":
        args[0], _, _ = evalVarString(args[0], keyValuesArray, context)
        if not isinstance(args[0], bool):
            error(context.added("lineNo", textIt.lineNo()),
                  "Argument to \\if must be a boolean")
        # Compare with True so that non-booleans evaluate to false
        if args[0] == True:
            useThisCase = True
        else:
            useThisCase = False
        if cmd == "\\ifnot":
            useThisCase = not useThisCase
        
    ifLines = []
    while not textIt.isDone():
        line = textIt.nextLine()
        match = kCmdRE.search(line)
        if match is None:  # normal line
            if useThisCase:
               ifLines.append(line)
        elif match.group(1) == "\\else":
            useThisCase = not useThisCase
        elif match.group(1) == "\\endif":
            break
        else:  # treat as normal line
            if useThisCase:
               ifLines.append(line)

    return ifLines

def readForLines(textIt, line, cmd, templates, keyValuesArray, context):
    nFor = 1
    forLines = []
    while not textIt.isDone():
        line = textIt.nextLine()
        match = kCmdRE.search(line)
        if match is None:
            forLines.append(line)
        else:
            if match.group(1) == "\\endfor":
                nFor -= 1
                if nFor == 0:
                    break
            elif match.group(1) == "\\for":
                nFor += 1
            forLines.append(line)   # treat as normal line
    return forLines

def expandVariables(templateText, templates, keyValues, extraKeyValues, context, ignoreErrors=False):
    # Iterate through matches from bottom up, so that replacing one match
    # doesn't invalidate all the others
    matches = sorted(kVariableFuncRE.finditer(templateText), reverse=True,
                     key=lambda m: m.start())
    # debugging
#    for m in matches:
#        print m.group(1), m.start(), m.end()
#    print keyValues.data
                     
    for match in matches:
        varStr = match.group(0)
        parenIdx = 0
        while parenIdx >= 0 and parenIdx < len(varStr):
            parenIdx = varStr.find('(', parenIdx)
            if parenIdx >= 0:
                closeParenIdx = varStr.find(')', parenIdx + 1)
                if varStr[parenIdx:closeParenIdx].find('\\'):
                    t = expandVariables(varStr[parenIdx + 1:closeParenIdx],
                                        templates, keyValues,
                                        extraKeyValues, context, ignoreErrors)
                    varStr = varStr[:parenIdx + 1] + t + varStr[closeParenIdx:]
                    parenIdx = varStr.find(')', parenIdx + 1) + 1

        v, typename, wasError = evalVarString(varStr,
                                              [keyValues, extraKeyValues],
                                              context, ignoreErrors=ignoreErrors)
        if ignoreErrors and wasError:
            if varStr != match.group(0):
                templateText = templateText[:match.start()] + varStr + templateText[match.end():]
            continue;

        # TODO:  don't resubstitute if data[k] has any \text
        if isinstance(v, ArrayValue):
            if typename in templates:   # currently, arrays don't have types
                templ = templates[typename]
            else:
                templ = Template("dummy.template", "[dummy template]", 0, "")
            t = evalArray(templates, templ, v, context, ignoreErrors)
            templateText = templateText[:match.start()] + t + templateText[match.end():]
#        elif isinstance(v, StructValue) or typename in templates:
        elif isinstance(v, StructValue):
            if typename in templates:
                t = instantiateTemplate(templates,
                                        templates[typename],
                                        v,
                                        context.added("template",
                                                      typename),
                                        globalVals=extraKeyValues)
                templateText = templateText[:match.start()] + t + templateText[match.end():]
            else:
                if not ignoreErrors:
                    error(context, str(typename) + " is not a template")
                    templateText = typename.upper()
        else:
            t = v.data
            if v.needsExpand():
                t = expandVariables(t, templates, keyValues, extraKeyValues,
                                    context, ignoreErrors)
            templateText = templateText[:match.start()] + t + templateText[match.end():]

    return templateText

def evalVarString(s, keyValuesArray, context, ignoreErrors=False):
    wasError = False
    s = s.replace("{", "").replace("}", "")
    varParts = s.split(".")
    k = varParts[0]
    filters = varParts[1:]
    typename = None
    v = None
    for kv in keyValuesArray:
        if kv is None:
            continue
        if k in kv.data:
            typename = kv.data[k].typename
            v = kv.data[k]
        elif k == kv.typename:
            typename = k
            v = kv
    if typename is not None:
        pass   # everything ok, we really want to handle when it is None
    # Key does not match anything.  Check if it is an common escape sequence
    # (e.g. if \n doesn't match anything, we are probably parsing Javascript
    # or something)
    elif k == "\\n":
        typename = kTypeString
        v = StringValue(typename, k, context)
    # If it is an eval, but we can't find the key, it might be that it is
    # nested in a data object, so just put it back and it should substitute
    # fine the next time.
    elif k.startswith(kEvalPrefix):
        typename = k
        v = StringValue(kTypeString, k, context)
    # Ok, this genuinely doesn't match anything
    else:
        typename = kTypeString
        v = StringValue(kTypeString, k.upper(), context)
        wasError = True
        if not ignoreErrors:
            error(context, "Key '" + k + "' required by template '" + context.data['template'] + "' is not present.")

    if len(filters) > 0 and not (wasError and ignoreErrors):
        v = evalFilters(v, filters, context.added("key", k))
        if isinstance(v, Value):
            typename = v.typename
        else:
            typename = None

    return v, typename, wasError

def evalFilters(value, filters, context):
    for f in filters:
        parenIdx = f.find("(")
        if parenIdx == 0:
            error(context, "Template arguments not supported yet")
        elif parenIdx > 0:
            func = f[:parenIdx]
            args = f[parenIdx + 1:-1]
            newValue = value.call(context, func, args)
            if newValue is not None:
                value = newValue
        else:  # no parens:  is struct reference
            f = "\\" + f
            if not isinstance(value, StructValue):
                error(context, "Key '" + context.data['key'] + "' must be a dictionary.")
                break
            elif f not in value.data:
                error(context, "Key '" + context.data['key'] + "' does not have '" + f + "'.")
                break
            else:
                value = value.data[f]
    return value

def evalArray(templates, arrayTemplate, array, context, ignoreErrors):
    t = ""
    idx = 0
    thisVal = ThisValue(context)
    thisVal.setLoopCount(len(array.data))
    globalVals = StructValue("", { kThisKey: thisVal }, context)
    for item in array.data:
        if item.typename in templates:
            keysWithArrayKeys = StructValue("", item.data, context)
            keysWithArrayKeys.data[kCurrentIndexKey] = StringValue(kTypeString, str(idx), context)
            if idx % 2 == 0:
                keysWithArrayKeys.data[kCurrentIndexIsEvenKey] = StringValue(kTypeString, '', context)
            else:
                keysWithArrayKeys.data[kCurrentIndexIsOddKey] = StringValue(kTypeString, '', context)
            templ = templates[item.typename] # item might be different type
                                             # each iteration
            # Note:  this will fail if not all items in the array create new
            #        files.
            if arrayTemplate.useNewFileForTemplate():
#                prefix = os.path.dirname(context.data["filename"])
#                if len(prefix) > 0:
#                    prefix += "/"
#                p = prefix + arrayTemplate.newFilePrefix() + str(idx) + ".html"
#                n = prefix + arrayTemplate.newFilePrefix() + str(idx+2) + ".html"
                # Use basename, because if the prefix contains a directory,
                # this file will be in that directory, so next/prev need to
                # be relative to that.
                prefix = os.path.basename(arrayTemplate.newFilePrefix())
                p = prefix + str(idx) + ".html"
                n = prefix + str(idx+2) + ".html"
                thisVal.setURLs(p, n)

            text = instantiateTemplate(templates, templ,
                                       keysWithArrayKeys,
                                       context.added("template",
                                                     item.typename),
                                       globalVals=globalVals)
            if arrayTemplate.useNewFileForTemplate():
                path = os.path.dirname(context.data["filename"])
                if len(path) > 0:
                    path += "/"
                path += arrayTemplate.newFilePrefix() + str(idx + 1) + ".html"
                text = beautifyHTML(text)
                writeFile(path, text, context.added("template", item.typename))
            else:
                t += text
            thisVal.incrementId()
        elif isinstance(item, StringValue):
            # TODO:  we should probably expand variables here, but we
            #        haven't been given that information.
            t += item.data
            thisVal.incrementId()
        else:
            if not ignoreErrors:
                error(context, str(item.typename) + " is not a template")
        idx += 1
    return t

def parseData(context, textIt, data, line=None, substInfo=None):
    newdataStartLineNo = 0
    newdata = {}
    def addKeyValue(k, v):
        if k is not None and k not in newdata:
            newdata[k] = v
            if v == "{":
                raise Exception("debug")
            if (isinstance(v, list) or isinstance(v, tuple)):
                if v[0] == kTypeString and not isinstance(v[1], str):
                    raise Exception("assert")
                if v[0] == kTypeArray and not isinstance(v[1], list):
                    raise Exception("assert")
                if v[0][0] == '\\' and not isinstance(v[1], dict):
                    raise Exception("assert")

    name = None
    key = None
    typename = kTypeString
    value = ""
    valueNeedsExpand = False
    nErrors = 0
    if line is None:
        line = textIt.nextLine()
    while line is not None:
        if kDataCloseRE.search(line) is not None:
            v = evaluateVariablesInData(value, key, substInfo,
                                        context.added("lineNo", textIt.lineNo()))
            if valueNeedsExpand:
                v.setNeedsExpand(True)
            addKeyValue(key, v)
            return (name, StructValue(name, newdata, context), nErrors)

        # Allow quoting of closing character (handy for writing code)
        if kBeginCodeRE.search(line) is not None:
            before, codeStr, line = readCodeLines(context, textIt, line)
            value += before
            value += codeStr

        if line.startswith("\\include "):
            inclFile = kIncludeRE.search(line).group(1)
            line = textIt.includeFile(inclFile, context.data["cwd"], context)

        isEval = False
        parseLineResults = kDataLineRE.search(line)
        if parseLineResults is not None and parseLineResults.group(1) in kKeywords:
            if parseLineResults.group(1) == '\\eval':
                isEval = True
            parseLineResults = None
        if parseLineResults is None and isEval == False:
            if name is None:
                error(context.added("lineNo", textIt.lineNo()),
                      "expecting: \\key [\\type] {")
                nErrors += 1
            else:
                if line.startswith('\\'):
                    match = kCmdRE.search(line)
                    if match is not None and match.group(1) in kKeywords:
                        line = '\n' + line + '\n'
                if value.find('\n') < 0:
                    value += '\n'
                value += line
        elif isEval == True:
            idx = line.find('\\eval')
            idx += 5  # increment past \eval
            while idx < len(line) and line[idx].isspace():
                idx += 1
            lineNo = textIt.lineNo()
            if line.find('{') >= 0:
                newdataStartLineNo = lineNo
                k, v, nerr = parseData(context, textIt, data, line=line[idx:],
                                       substInfo=substInfo)
            elif line.find('[') >= 0:
                v, nerr = parseArray(context, textIt, data,
                                     substInfo=substInfo)
                v = ArrayValue(v, context.added("lineNo", lineNo))
            else:
                nerr = 1
                error(context.added("lineNo", lineNo),
                      "\\eval requires struct or array")
            nErrors += nerr
            fakeKey = kEvalPrefix + str(lineNo)
            if nerr == 0:
                addKeyValue(fakeKey, v)
                value += "\n" + fakeKey + "\n"
                valueNeedsExpand = True;
        else:
            v = evaluateVariablesInData(value, key, substInfo,
                                        context.added("lineNo", textIt.lineNo()))
            if valueNeedsExpand:
                v.setNeedsExpand(True)
            addKeyValue(key, v)
            value = parseLineResults.group(2)
            key = parseLineResults.group(1)
            typeResults = kDataLineRE.search(value)
            if typeResults is None:
                typename = kTypeString
            else:
                typename = typeResults.group(1)
                value = typeResults.group(2)
            if len(value) > 0 and value[0] == '{':
                if len(value) > 1 and not value[1:].isspace():
                    error(context.added("lineNo", textIt.lineNo()),
                          "expecting: '\\key [\\type] {';  found '" + value + " " + value + "'")
                    nErrors += 1
                if name is None:
                    name = parseLineResults.group(1)
                else:
                    k, v, nerr = parseData(context, textIt, data, line=line,
                                           substInfo=substInfo)
                    nErrors += nerr
                    if k not in newdata:
                        newdata[k] = v
                key = None
                value = ""
            elif len(value) > 0 and value[0] == '[':
                lineNo = textIt.lineNo()
                v, nerr = parseArray(context, textIt, data, substInfo=substInfo)
                if key not in newdata:
                    newdata[key] = ArrayValue(v, context.added("lineNo", lineNo))
                key = None
                value = ""
            else:
                if name is None:
                    error(context.added("lineNo", textIt.lineNo()),
                          ": expecting: \\key [\\type] {")
                    nErrors += 1
                
                if key in data:
                    error(context.added("lineNo", textIt.lineNo()),
                          "key '" + key + "' already seen.")
                    nErrors += 1
                # Note:  delay putting key in until we have read all the lines...
            
        line = textIt.nextLine()

    # Check if the data is a giant array (or maybe string), instead of a struct
    if len(newdata) == 1 and not isinstance(newdata.values()[0], StructValue):
        return newdata.keys()[0], newdata.values()[0], nErrors
        
    error(context.added("lineNo", textIt.lineNo()),
          "expecting closing } at end of file")
    nErrors += 1

    return key, StructValue(typename, newdata,
                            context.added("lineNo", newdataStartLineNo)), nErrors

def parseArray(context, textIt, data, line=None, substInfo=None):
    nErrors = 0
    array = []
    startLineNo = textIt.lineNo()
    key = None
    value = ""
    if line is None:
        line = textIt.nextLine()
    while line is not None:
        if line.startswith("\\include "):
            inclFile = kIncludeRE.search(line).group(1)
            line = textIt.includeFile(inclFile, context.data["cwd"], context)
        if kArrayCloseRE.search(line) is not None:
            if key is not None:
                v = evaluateVariablesInData(value, key, substInfo,
                                          context.added("lineNo", textIt.lineNo()))
                array.append(v)
            return (array, nErrors)
        lineResults = kDataLineRE.search(line)
        if lineResults is None:
            value += line
        else:
            if key is not None:
                v = evaluateVariablesInData(value, key, substInfo,
                                          context.added("lineNo", textIt.lineNo()))
                array.append(v)
            value = lineResults.group(2)
            if len(value) > 0 and value[0] == '{':
                k, v, nerr = parseData(context, textIt, data, line, substInfo)
                array.append(v)
                nErrors += nerr
                key = None
                value = ""
            else:
                key = lineResults.group(1)
        line = textIt.nextLine()

    error(context.added("lineNo", textIt.lineNo()),
          "array that started at line " + str(startLineNo) + " did not end")
    return (array, nErrors + 1)

def evaluateVariablesInData(text, typename, substInfo, context):
    if not substInfo:
        return StringValue(typename, text, context)
    if len(text) == 0:
        return StringValue(typename, text, context)

    # Check if entire text is a variable, in which case substitute it whole
    if substInfo["keyValues"] is not None and text in substInfo["keyValues"].data:
        return substInfo["keyValues"].data[text]
    if substInfo["globalVals"] is not None and text in substInfo["globalVals"].data:
        return substInfo["globalVals"].data[text]

    # Otherwise, substitute the text
    text = instantiateText(TextIterator(text),
                           os.path.dirname(context.data["filename"]),
                           substInfo["templates"], substInfo["keyValues"],
                           substInfo["globalVals"], context)

    return StringValue(typename, text, context)

def readCodeLines(context, textIt, line):
    match = kBeginCodeRE.search(line)
    if match is None:
        return (line, "", "")
    start = match.start(0)
    before = line[:start]
    line = line[match.end(0) + 1:]
    while len(line) > 0 and line[0].isspace():
        line = line[1:]
    codeStr = ""
    while line is not None:
# `line` already has \n at the end
#        if len(codeStr) > 0:
#            codeStr += "\n"

        match = kEndCodeRE.search(line)
        if match is not None:
            codeStr += line[:match.start(0)]
            line = line[match.end(0) + 1:]
            break
        
        codeStr += line
        line = textIt.nextLine()

    return (before, codeStr, line)
    
def searchForFile(filename, pathPrefix):
    path = filename
    if not (path.startswith('./') or path.startswith('../') or
            path.startswith('/')) and len(pathPrefix) > 0:
        path = pathPrefix + '/' + filename
    if os.path.isfile(path):
        return path

    for d in gIncludeDirs:
        path = d + '/' + path
        if os.path.isfile(path):
            return path

    return filename

def calcHTMLFilename(path):
    dotIdx = path.rfind(".")
    if dotIdx < 0:
        return path + ".html"
    else:
        return path[:dotIdx] + ".html"
        
def calcRelativePath(cwd, path):
    idx = len(cwd)
    while idx < len(path) and path[idx] == '/':
        idx += 1
    return path[idx:]

def writeFile(outpath, text, errorContext):
    if gIsVerbose:
        print "writing:", outpath

    # Make the directory, in case it isn't there
    # If directory is '', this is equivalent to './', which we know exists
    d = os.path.dirname(outpath)
    if len(d) != 0:
        try:
            os.makedirs(d)
        except OSError as e:  # Python >2.5
            if e.errno == errno.EEXIST and os.path.isdir(d):
                pass
            else:
                raise e

    # Now write the file
    try:
        OUT = open(outpath, "w")
        print >>OUT, text
        OUT.close()
    except IOError, e:
        error(errorContext, "Error writing output to '" + outpath + "': " + e.message)

#-------------------------------------------------------------------------------
class Template(object):
    def __init__(self, path, name, startLineNo, content):
        self._path = path
        self._name = name
        self._startLineNo = startLineNo
        self._content = content
        self._directory = ""
        self._arrayFilenamePrefix = None
        self._indexSubdirs = False
        idx = path.rfind('/')
        if idx >= 0:
            self._directory = path[:idx]

    def __str__(self):
        if len(self._content) > 30:
            s = self._content[:30] + "..."
        else:
            s = self._content
        s = s.replace('\n', '\\n')
        return s

    def __repr__(self):
        return self.__str__()

    def directory(self):
        return self._directory

    def path(self):
        return self._path

    def name(self):
        return self._name

    def lineNo(self):
        return self._startLineNo

    def content(self):
        return self._content

    def indexSubdirs(self):
        return self._indexSubdirs

    def setIndexSubdirs(self, index):
        self._indexSubdirs = index 

    def useNewFileForTemplate(self):
        return self._arrayFilenamePrefix is not None

    def setNewFilePrefix(self, prefix):
        self._arrayFilenamePrefix = prefix

    def newFilePrefix(self):
        return self._arrayFilenamePrefix

#-------------------------------------------------------------------------------
class Value(object):
    def __init__(self, typename, data, filename, lineNo):
        self.typename = typename
        self.data = data
        if isinstance(filename, Context):
            self.filename = filename.data["filename"]
            if "lineNo" in filename.data:
                self.lineNo = filename.data["lineNo"]
            else:
                self.lineNo = 0
        else:
            self.filename = filename
            self.lineNo = lineNo

    def __str__(self):
        return self.toString()

    def __repr__(self):
        return self.toString()

    def toString(self, indent=""):
        raise Exception("Value.toString() needs to be overridden")

    def contentAsTextLines(self, name="", indent=""):
        raise Exception("Value.toTemplateText() needs to be overridden")

    def call(self, context, f, argStr):
        # Split into args.  (re.split gives '' at beginning and end,
        # and also includes the separators, too much effort to remove)
        args = []
        startIdx = 0
        for idx in xrange(1, len(argStr)):
            if argStr[idx] == ',':
                args.append(argStr[startIdx:idx])
                startIdx = idx + 1
            elif argStr[idx].isspace():
                startIdx = idx + 1
        if startIdx < len(argStr):
            args.append(argStr[startIdx:])

        # Convert numbers into numbers
        for idx in xrange(0, len(args)):
            if args[idx].isdigit():
                args[idx] = float(args[idx])

        if not hasattr(self, f):
            error(context, "No function '" + f + "' for type " + f.__class__.__name__)
            return None

        if callable(getattr(self, f)):
            return getattr(self, f)(context, args)

    def verifyArgs(self, context, fname, argTypes, args):
        if len(argTypes) != len(args):
            error(context, fname + " expects " + str(len(argTypes)) + ", given " + str(len(args)))
            return False
        for idx in xrange(len(argTypes)):
            if not isinstance(args[idx], argTypes[idx]):
                need = ", ".join([t.__name__ for t in argTypes])
                got = ", ".join([t.__class__.__name__ for t in args])
                error(context, fname + " expects " + fname + "(" + need + "), given " + fname + "(" + got + ")")
                return False
        return True

    # url() string
    def url(self, context, args):
        if not self.verifyArgs(context, "url", [], args):
            return "url(ERROR)"
        return StringValue(kTypeString, calcHTMLFilename(calcRelativePath(context.data["cwd"], self.filename)), self.filename, self.lineNo)

    # isType() bool
    def isType(self, context, args):
        if not self.verifyArgs(context, "isType", [basestring], args):
            return StringValue(kTypeString, "isType(ERROR)", self.filename,
                               self.lineNo)
        return (self.typename == "\\" + args[0])

class StringValue(Value):
    def __init__(self, typename, data, filename, lineNo=0):
        super(StringValue, self).__init__(typename, data, filename, lineNo)
        self._needsExpand = False
        if not isinstance(data, basestring):
            raise Exception("Internal error:  StringValue requires a string;  got " + type(data).__name__)

    def toString(self, indent=""):
        if len(self.data) <= 60:
            return "(" + self.typename + ') "' + str(self.data) + '"'
        else:
            return "(" + self.typename + ') "' + str(self.data)[:61] + '..."'

    def contentAsTextLines(self, name="", indent=""):
        if name == "":
            return [ indent + self.typename + " " + str(self.data) ]
        else:
            return [ indent + name + " " + str(self.data) ]

    def needsExpand(self):
        return self._needsExpand

    def setNeedsExpand(self, needs):
        self._needsExpand = needs

    # replace(old, new) string
    def replace(self, context, args):
        if not self.verifyArgs(context, "replace", [basestring, basestring], args):
            return StringValue(kTypeString, "replace(ERROR)", context)
        return StringValue(kTypeString, self.data.replace(args[0], args[1]), context)

    # substr(start, len) string
    def substr(self, context, args):
        if not self.verifyArgs(context, "substr", [float, float], args):
            return StringValue(kTypeString, "substr(ERROR)", context)
        return StringValue(kTypeString, self.data[int(args[0]) : int(args[0])+int(args[1])], context)

    # toLower() string
    def toLower(self, context, args):
        if not self.verifyArgs(context, "toLower", [], args):
            return StringValue(kTypeString, "toLower(ERROR)", context)
        return StringValue(kTypeString, self.data.lower(), context)

    # toUpper() string
    def toUpper(self, context, args):
        if not self.verifyArgs(context, "toUpper", [], args):
            return StringValue(kTypeString, "toUpper(ERROR)", context)
        return StringValue(kTypeString, self.data.upper(), context)

    def urlEncode(self, context, args):
        if not self.verifyArgs(context, "urlEncode", [], args):
            return StringValue(kTypeString, "urlEncode(ERROR)", context)
        return StringValue(kTypeString, urllib.quote(self.data), context)

    # split(separator) string
    def split(self, context, args):
        if not self.verifyArgs(context, "split", [basestring], args):
            return ArrayValue([], context)
        results = self.data.split(args[0])
        results = [StringValue(kTypeString, s, context) for s in results]
        return ArrayValue(results, context)

    # equals(string) bool
    def equals(self, context, args):
        if len(args) >= 1 and not isinstance(args[0], basestring):
            args[0] = str(args[0])
        if not self.verifyArgs(context, "equals", [basestring], args):
            return False
        return self.data == args[0]

    # notEquals(string) bool
    def notEquals(self, context, args):
        if len(args) >= 1 and not isinstance(args[0], basestring):
            args[0] = str(args[0])
        if not self.verifyArgs(context, "notEquals", [basestring], args):
            return False
        return self.data != args[0]

    # startsWith(string) bool
    def startsWith(self, context, args):
        if not self.verifyArgs(context, "startsWith", [basestring], args):
            return False
        return self._data.startswith(args[0])

    # contains(string) bool
    def contains(self, context, args):
        if not self.verifyArgs(context, "contains", [basestring], args):
            return False
        needle = args[0]
        needle = needle.replace("\\n", "\n")
        needle = needle.replace("\\r", "\r")
        needle = needle.replace("\\t", "\t")
        return self.data.find(needle) >= 0

    # endsWith(string) bool
    def endsWith(self, context, args):
        if not self.verifyArgs(context, "endsWith", [basestring], args):
            return False
        return self._data.endswith(args[0])

    # add(float) string
    def add(self, context, args):
        if not self.verifyArgs(context, "add", [float], args):
            return self
        val = 0
        try:
            val = float(self.data)
        except ValueError:
            error(context, "add() can only be called on a string that is an number")
        return StringValue(self.typename, str(val + args[0]), self.filename,
                           self.lineNo)

    # subtract(float) string
    def subtract(self, context, args):
        if not self.verifyArgs(context, "subtract", [float], args):
            return self
        val = 0
        try:
            val = float(self.data)
        except ValueError:
            error(context, "subtract() can only be called on a string that is an number")
        return StringValue(self.typename, str(val + args[0]), self.filename,
                           self.lineNo)

    # multiply(float) string
    def multiply(self, context, args):
        if not self.verifyArgs(context, "multiply", [float], args):
            return self
        val = 0
        try:
            val = float(self.data)
        except ValueError:
            error(context, "multiply() can only be called on a string that is an number")
        return StringValue(self.typename, str(val + args[0]), self.filename,
                           self.lineNo)

    # divide(float) string
    def divide(self, context, args):
        if not self.verifyArgs(context, "divide", [float], args):
            return self
        val = 0
        try:
            val = float(self.data)
        except ValueError:
            error(context, "divide() can only be called on a string that is an number")
        return StringValue(self.typename, str(val + args[0]), self.filename,
                           self.lineNo)

    # year("YYYY[MM[DD]]") string
    def year(self, context, args):
        if len(args) == 1:
            if args[0].isdigit() and (len(args[0]) == 4 or len(args[0]) == 6 or len(args[0]) == 8):
                pass   # everything is okay!
            else:
                error(context, fname + " expects " + fname + '("YYYY[MM[DD]]"), given ' + fname + "(" + args[0]+ ")")
        else:
            error(context, fname + ' expects 1 argument ("YYYY[MM[DD]]"), given ' + str(len(args)))
        return StringValue(self.typename, args[0][0:4], self.filename,
                           self.lineNo)

    # month("YYYYMM[DD]") string      Returns one or two digit month
    def month(self, context, args):
        if len(args) == 1:
            if args[0].isdigit() and (len(args[0]) == 6 or len(args[0]) == 8):
                pass   # everything is okay!
            else:
                error(context, fname + " expects " + fname + '("YYYYMM[DD]"), given ' + fname + "(" + args[0]+ ")")
                while len(args[0]) < 6:
                    args[0] += "0"
        else:
            error(context, fname + ' expects 1 argument ("YYYYMM[DD]"), given ' + str(len(args)))
        return StringValue(self.typename, args[0][4:6], self.filename,
                           self.lineNo)

    # twoDigitMonth("YYYYMM[DD]") string      Returns two digit month
    def twoDigitMonth(self, context, args):
        if len(args) == 1:
            if args[0].isdigit() and (len(args[0]) == 6 or len(args[0]) == 8):
                pass   # everything is okay!
            else:
                error(context, fname + " expects " + fname + '("YYYYMM[DD]"), given ' + fname + "(" + args[0]+ ")")
                while len(args[0]) < 6:
                    args[0] += "0"
        else:
            error(context, fname + ' expects 1 argument ("YYYYMM[DD]"), given ' + str(len(args)))
        mon = args[0][4:6]
        while len(mon) < 2:
            mon += "0"
        return StringValue(self.typename, mon, self.filename, self.lineNo)

    # monthStr("YYYYMM[DD]") string   Returns long name of month
    def monthStr(self, context, args):
        if len(args) == 1:
            if args[0].isdigit() and (len(args[0]) == 6 or len(args[0]) == 8):
                pass   # everything is okay!
            else:
                error(context, fname + " expects " + fname + '("YYYYMM[DD]"), given ' + fname + "(" + args[0]+ ")")
                while len(args[0]) < 6:
                    args[0] += "0"
        else:
            error(context, fname + ' expects 1 argument ("YYYYMM[DD]"), given ' + str(len(args)))
        try:
            mon = int(args[0][4:6]) - 1
            if mon < 0 or mon > 11:
                error(context, fname + ' month must be between 1 and 12.')
                return StringValue(self.typename, "MONTH", self.filename)
            return StringValue(self.typename,
                               [ "January", "February", "March", "April", "May",
                                 "June", "July", "August", "September",
                                 "October", "November", "December" ][mon],
                               self.filename, self.lineNo)
        except ValueError:
            return StringValue(self.typename, "MONTH", self.filename,
                               self.lineNo)

    # day("YYYYMMDD") string      Returns one or two digit day
    def day(self, context, args):
        if len(args) == 1:
            if args[0].isdigit() and len(args[0]) == 8:
                pass   # everything is okay!
            else:
                error(context, fname + " expects " + fname + '("YYYYMMDD"), given ' + fname + "(" + args[0]+ ")")
                while len(args[0]) < 8:
                    args[0] += "0"
        else:
            error(context, fname + ' expects 1 argument ("YYYYMMDD"), given ' + str(len(args)))
        return StringValue(self.typename, args[0][6:8], self.filename,
                           self.lineNo)

    # twoDigitDay("YYYYMMDD") string      Returns two digit day
    def twoDigitDay(self, context, args):
        if len(args) == 1:
            if args[0].isdigit() and (len(args[0]) == 6 or len(args[0]) == 8):
                pass   # everything is okay!
            else:
                error(context, fname + " expects " + fname + '("YYYYMMDD"), given ' + fname + "(" + args[0]+ ")")
                while len(args[0]) < 8:
                    args[0] += "0"
        else:
            error(context, fname + ' expects 1 argument ("YYYYMMDD"), given ' + str(len(args)))
        day = args[0][6:8]
        while len(day) < 2:
            day += "0"
        return StringValue(self.typename, day, self.filename, self.lineNo)

    # codeToHTML(language) string
    # Returns code pretty-printed in HTML based on the language specified
    def codeToHTML(self, context, args):
        if len(args) != 1:
            error(context, 'pythonToHTML() expects 1 argument, given ' + str(len(args)))
        # Remove any empty lines at the begining.  It usually comes from doing:
        # \eval \code {
        #   \python
        # ... code here ...
        # }
        # (obviously not limited to \python)
        # This is a usability feature, since it is awkward to have to put your
        # first line of code on the language line and then it's indented wrong.
        code = self.data
        while len(code) > 0 and (code[0] == '\n' or code[0] == '\r'):
            code = code[1:]

        # Remove any trailing empty lines, too.  This is a usability feature
        # to avoid having to figure out why the bottom margin looks wrong.
        while len(code) > 0 and (code[-1] == '\n' or code[-1] == '\r'):
            code = code[:-1]

        parser = CodeToHTML(args[0])
        htmlCode = parser.parse(code)
        return StringValue(self.typename, htmlCode, self.filename, self.lineNo)

class ArrayValue(Value):
    def __init__(self, data, filename, lineNo=0):
        super(ArrayValue, self).__init__(kTypeArray, data, filename, lineNo)
        if not isinstance(data, list):
            raise Exception("Internal error:  ArrayValue requires a array")
        if len(data) > 0 and isinstance(data[0], basestring):
            raise Exception("Internal error:  Can't put [str,...] into ArrayValue, need [StringValue,...]")

    def toString(self, indent=""):
        s = indent + "[\n"
        newindent = indent + "  "
        for e in self.data:
            s += newindent + e.toString(newindent) + "\n"
        s += indent + "]"
        return s

    def contentAsTextLines(self, name="", indent=""):
        lines = []
        if name != "":
            lines.append(indent + name + " [")
        newindent = indent + "  "
        for e in self.data:
            lines += e.contentAsTextLines(e.typename, newindent)
        if name != "":
            lines.append(indent + "]")
        return lines

    def elementAt(self, context, args):
        if not self.verifyArgs(context, "elementAt", [float], args):
            return StringValue(kTypeString, "elementAt(ERROR)", context)
        idx = int(args[0])
        if idx < len(self.data):
            return self.data[idx]
        else:
            error(context, "elementAt(" + str(idx) + "):  index out of bounds;  array has length " + str(len(self.data)))
            return StringValue(kTypeString, "", context)

    def sortBy(self, context, args):
        if not self.verifyArgs(context, "sortBy", [basestring], args):
            results = []
        else:
            results = self.data
#            nKeyErrors = 0
#            nTypeErrors = 0
            key = "\\" + args[0]
            def compareBy(o):
                if key in o.data:
                    v = o.data[key]
                    if isinstance(v.data, basestring):
                        return v.data
                    else:
#                        nTypeErrors += 1
                        return v.toString()
                else:
#                    nKeyErrors += 1
                    return ""
            results = sorted(self.data, key=compareBy)
#            if nKeyErrors > 0:
#                error(context, "sortBy: " + str(nKeyErrors) + " elements in array did not have '" + key + "'.")
#            if nTypeErrors > 0:
#                error(context, "sortBy: " + str(nTypeErrors) + " elements in array were not strings.")
        return ArrayValue(results, context)

    def reverseSortBy(self, context, args):
        if not self.verifyArgs(context, "reverseSortBy", [basestring], args):
            return []
        av = self.sortBy(context, args)
        av.data.reverse()
        return av

    def sortByTitle(self, context, args):
        if not self.verifyArgs(context, "sortByTitle", [basestring], args):
            results = []
        else:
            results = self.data
            key = "\\" + args[0]
            def compareBy(o):
                if key in o.data:
                    v = o.data[key]
                    if isinstance(v.data, basestring):
                        v = v.data
                    else:
                        v = v.toString()
                else:
                    v = ""
                if v.startswith("The ") or v.startswith("the "):
                    v = v[4:] + ", The"
                elif v.startswith("A ") or v.startswith("a "):
                    v = v[2:] + ", A"
                return v
            results = sorted(self.data, key=compareBy)
        return ArrayValue(results, context)


class StructValue(Value):
    def __init__(self, typename, data, filename, lineNo=0):
        super(StructValue, self).__init__(typename, data, filename, lineNo)
        if not isinstance(data, dict):
            raise Exception("Internal error:  StructValue requires a dict")
#        if typename == kTypeString:
#            raise Exception("Internal error:  StructValue typename cannot be '_string'")

    def toString(self, indent=""):
        s = indent + "{\n"
        newindent = indent + "  "
        for k,v in self.data.iteritems():
            s += newindent + k + ": " + v.toString(newindent) + "\n"
        s += indent + "}"
        return s

    def contentAsTextLines(self, name="", indent=""):
        lines = []
        if name != "":
            lines.append(indent + name + " [");
        newindent = indent + "  "
        for k,v in self.data.iteritems():
            lines += v.contentAsTextLine(k, newindent)
        if name != "":
            lines.append(indent + "]");
        return s

class ThisValue(StructValue):
    _nextGlobalId = 0
    def __init__(self, filename, lineNo=0):
        super(ThisValue, self).__init__(kTypeThis, {}, filename, lineNo)
        self._nextId = 1
        self._loopCount = None
        self._nexturl = None
        self._prevurl = None

    def toString(self, indent=""):
        return self.typename + ": nextId " + str(self._nextId)

    def incrementId(self):
        self._nextId += 1

    def setLoopCount(self, n):
        self._loopCount = n

    def setURLs(self, p, n):
        self._nexturl = n
        self._prevurl = p

    # nextGlobalId() string
    def nextGlobalId(self, context, args):
        if not self.verifyArgs(context, "nextGlobalId", [], args):
            return StringValue(kTypeString, "nextGlobalId(ERROR)", self.filename,
                               self.lineNo)
        ThisValue._nextGlobalId += 1
        return StringValue(kTypeString, str(ThisValue._nextGlobalId),
                           self.filename, self.lineNo)

    # idFor(basetext) string
    def idFor(self, context, args):
        if not self.verifyArgs(context, "idFor", [basestring], args):
            return StringValue(kTypeString, "idFor(ERROR)", self.filename,
                               self.lineNo)
#        if args[0] not in self.data:
#            self.data[args[0]] = self._nextId
        return StringValue(kTypeString, args[0] + str(self._nextId),
                           self.filename, self.lineNo)

    # loopIndex() string,  1-indexed
    def loopIndex(self, context, args):
        if not self.verifyArgs(context, "loopIndex", [], args):
            return StringValue(kTypeString, "loopIndex(ERROR)", self.filename,
                               self.lineNo)
        return StringValue(kTypeString, str(self._nextId), self.filename,
                           self.lineNo)

    # loopCount() string
    def loopCount(self, context, args):
        if self._loopCount is None:
            error(context, "loopCount() can only be used with arrays")
            return StringValue(kTypeString, "0", self.filename)
        if not self.verifyArgs(context, "loopIndex", [], args):
            return StringValue(kTypeString, "loopIndex(ERROR)", self.filename,
                               self.lineNo)
        return StringValue(kTypeString, str(self._loopCount), self.filename,
                               self.lineNo)

    # isLoopStart() bool
    def isLoopStart(self, context, args):
        if not self.verifyArgs(context, "isLoopStart", [], args):
            return False
        return self._nextId <= 1

    # isLoopEnd() bool
    def isLoopEnd(self, context, args):
        if self._loopCount is None:
            error(context, "isLoopEnd() can only be used with arrays")
            return True
        if not self.verifyArgs(context, "isLoopEnd", [], args):
            return StringValue(kTypeString, "isLoopEnd(ERROR)", self.filename,
                               self.lineNo)
        return self._nextId >= self._loopCount

    # nextURL() string
    def nextURL(self, context, args):
        if not self.verifyArgs(context, "nextURL", [], args):
            return StringValue(kTypeString, "nextURL(ERROR)", self.filename,
                               self.lineNo)
        if self._nexturl is None:
            error(context, "nextURL() can only be used with an array")
            return StringValue(kTypeString, "ERROR", self.filename,
                               self.lineNo)
        return StringValue(kTypeString, self._nexturl, self.filename,
                               self.lineNo)

    # prevURL() string
    def prevURL(self, context, args):
        if not self.verifyArgs(context, "prevURL", [], args):
            return StringValue(kTypeString, "prevURL(ERROR)", self.filename,
                               self.lineNo)
        if self._prevurl is None:
            error(context, "prevURL() can only be used with an array")
            return StringValue(kTypeString, "ERROR", self.filename,
                               self.lineNo)
        return StringValue(kTypeString, self._prevurl, self.filename,
                               self.lineNo)
        
#class CodeValue(Value):
#    def __init__(self, data, filename):
#        super(CodeValue, self).__init__("__code", data, filename)
#        if not isinstance(data, basestring):
#            raise Exception("Internal error:  CodeValue requires a string")
#
#    def toString(self, indent=""):
#        return self.data

#-------------------------------------------------------------------------------
class TextIterator:
    def __init__(self, s):
        if isinstance(s, basestring):
            self._lines = s.splitlines()
        elif isinstance(s, list):
            self._lines = s
        else:
            raise Exception("TextIterator must be called with a string or a list of strings, not " + s.__class__.__name__)
        self._nextIdx = 0

    def isDone(self):
        return self._nextIdx >= len(self._lines)

    def nextLine(self):
        if self._nextIdx < len(self._lines):
            self._nextIdx += 1
            return self._lines[self._nextIdx - 1]
        return None

    def prevLine(self):
        if self._nextIdx > 0:
            self._nextIdx -= 1
            return self._lines[self._nextIdx + 1]
        return None

    def lineNo(self):
        return self._nextIdx   # start at 1, so _nextIdx is right

    def save(self):
        return self._nextIdx

    def restore(self, saved):
        self._nextIdx = saved

    # Returns new current line
    def includeFile(self, path, pathPrefix, context):
        inclFile = searchForFile(path, pathPrefix)
        try:
            IN = open(inclFile)
            inclLines = IN.readlines()
            IN.close()
            if self._nextIdx == 0:
                self._lines = inclLines + self._lines[1:]
                return None
            else:
                self._lines = self._lines[0:self._nextIdx-1] + inclLines + self._lines[self._nextIdx:]
                return self._lines[self._nextIdx-1]
        except IOError, e:
            error(context.added("lineNo", self._nextIdx + 1), "Error including file '" + inclFile + "': " + e.message)
            return self._lines[self._nextIdx-1]

class CodeToHTML:
    C_COMMENTS = 1      # // and /* ... */
    POUND_COMMENTS = 2  # like bash and python comments (like this one)
    STD_KEYWORDS = [ "for", "while", "do", "if", "else", "break", "continue",
                     "return" ]
    C_KEYWORDS = [ "auto", "break", "case", "continue", "default", "do", "else",
                   "enum", "extern", "for", "goto", "if", "register", "return",
                   "sizeof", "static", "struct", "switch", "typedef", "union",
                   "volatile", "while" ]
    C_TYPES = [ "char", "const", "double", "float", "int", "long", "short",
                "signed", "unsigned", "void" ]

    CLASS_COMMENT = "codeComment"
    CLASS_KEYWORD = "codeKeyword"
    CLASS_TYPE = "codeType"
    CLASS_CONSTANT = "codeConstant"
    CLASS_FUNCTION_CALL = "codeFunctionCall"

    def __init__(self, language):
        if language == "python":
            self.keywordSet = frozenset(keyword.kwlist + dir(__builtins__))
            self.functionSet = frozenset(dir(__builtins__))
            self.typeSet = frozenset()
            self.constantSet = frozenset(["False", "True"])
            self.commentType = CodeToHTML.POUND_COMMENTS
        elif language == "c":
            self.keywordSet = frozenset(CodeToHTML.C_KEYWORDS)
            self.functionSet = frozenset()
            self.typeSet = frozenset(CodeToHTML.C_TYPES)
            self.commentType = CodeToHTML.C_COMMENTS
        elif language == "c++" or language == "cplusplus":
            self.keywordSet = frozenset(CodeToHTML.C_KEYWORDS +
                [ "asm", "catch", "class", "compl", "consteval", "constexpr",
                  "const_cast", "decltype", "delete", "dynamic_cast",
                  "explicit", "export", "friend", "inline", "mutable",
                  "namespace", "new", "noexcept", "operator", "private",
                  "protected", "public", "reinterpret_cast", "requires",
                  "static_assert", "static_cast", "template", "this",
                  "thread_local", "throw", "try", "typeid", "typename", "using",
                  "virtual", "volatile" ])
            self.functionSet = frozenset()
            self.typeSet = frozenset(CodeToHTML.C_TYPES +
                [ "auto", "bool", "char8_t", "char16_t", "char32_t", "uint32_t",
                  "size_t", "wchar_t",
                  "vector", "map", "set", "unordered_map", "unordered_set",
                  "list", "queue" ])
            self.constantSet = frozenset(["nullptr", "false", "true"])
            self.commentType = CodeToHTML.C_COMMENTS
        elif language == "javascript":
            self.keywordSet = frozenset(
                [ "break", "case", "catch", "class", "const", "continue",
                  "debugger", "default", "delete", "do", "else", "export",
                  "extends", "finally", "for", "function", "if", "import",
                  "in", "instanceof", "new", "return", "super", "switch", "this",
                  "throw", "try", "typeof", "var", "void", "while", "with",
                  "yield" ] +
                [ "enum", "implements", "interface", "let", "package", "private",
                  "protected", "public", "static", "await" ])
            self.functionSet = frozenset()
            self.typeSet = frozenset()
            self.constantSet = frozenset(["null", "false", "true"])
            self.commentType = CodeToHTML.C_COMMENTS
        elif language == "glsl":
            self.keywordSet = frozenset(CodeToHTML.C_KEYWORDS +
                [ "attribute", "uniform", "varying", "layout", "centroid",
                  "flat", "smooth", "noperspective", "patch", "sample",
                  "subroutine", "in", "out", "inout", "invariant", "discard",
                  "lowp", "mediump", "highp", "precision" ])
            self.functionSet = frozenset(
                [ "radians", "degrees", "sin", "cos", "tan", "asin", "acos",
                  "atan", "pow", "exp", "exp2", "log", "log2", "sqrt",
                  "inversesqrt", "abs", "sign", "floor", "ceil", "fract", "mod",
                  "min", "max", "clamp", "mix", "step", "smoothstep", "length",
                  "distance", "dot", "cross", "normalize", "faceforward",
                  "reflect", "refract", "matrixCompMult", "lessThan",
                  "lessThanEqual", "greaterThan", "greaterThanEqual", "equal",
                  "notEqual", "any", "all", "not", "texture2D", "textureCube" ])
            self.typeSet = frozenset(CodeToHTML.C_TYPES + ["bool"] +
                [ "mat2", "mat3", "mat4", "dmat2", "dmat3", "dmat4", "mat2x2",
                  "mat2x3", "mat2x4", "dmat2x2", "dmat2x3", "dmat2x4", "mat3x2",
                  "mat3x3", "mat3x4", "dmat3x2", "dmat3x3", "dmat3x4", "mat4x2",
                  "mat4x3", "mat4x4", "dmat4x2", "dmat4x3", "dmat4x4", "vec2",
                  "vec3", "vec4", "ivec2", "ivec3", "ivec4", "bvec2", "bvec3",
                  "bvec4", "dvec2", "dvec3", "dvec4", "uvec2", "uvec3", "uvec4",
                  "sampler1D", "sampler2D", "sampler3D", "samplerCube",
                  "sampler1DShadow", "sampler2DShadow", "samplerCubeShadow",
                  "sampler1DArray", "sampler2DArray", "sampler1DArrayShadow",
                  "sampler2DArrayShadow", "isampler1D", "isampler2D",
                  "isampler3D", "isamplerCube", "isampler1DArray",
                  "isampler2DArray", "usampler1D", "usampler2D", "usampler3D",
                  "usamplerCube", "usampler1DArray", "usampler2DArray",
                  "sampler2DRect", "sampler2DRectShadow", "isampler2DRect",
                  "usampler2DRect", "samplerBuffer", "isamplerBuffer",
                  "usamplerBuffer", "sampler2DMS", "isampler2DMS",
                  "usampler2DMS", "sampler2DMSArray", "isampler2DMSArray",
                  "usampler2DMSArray", "samplerCubeArray",
                  "samplerCubeArrayShadow", "isamplerCubeArray",
                  "usamplerCubeArray" ])
            self.constantSet = frozenset(["false", "true"])
            self.commentType = CodeToHTML.C_COMMENTS
        elif language == "xml":
            self.parse = self.parseXML
        else:
            self.keywordSet = frozenset(CodeToHTML.STD_KEYWORDS)
            self.functionSet = frozenset()
            self.commentType = CodeToHTML.C_COMMENTS

    def parse(self, code):
        startIdx = 0
        endIdx = 0

        def subst(code, clazz, startIdx, endIdx):
            newstr = '<span class="' + clazz + '">' + code[startIdx:endIdx] + '</span>'
            code = code[0:startIdx] + newstr + code[endIdx:]
            startIdx = startIdx + len(newstr)
            return code, startIdx

        def tokenEnd(code, idx, isToken):
            while idx < len(code) and isToken(code[idx]):
                idx += 1
            return idx

        while startIdx < len(code) and (code[startIdx] == '\n' or code[startIdx] == '\r'):
            startIdx += 1

        while startIdx < len(code):
            endIdx = startIdx
            c = code[startIdx]
            if ((c == "#" and self.commentType == CodeToHTML.POUND_COMMENTS) or
                (c == "/" and self.commentType == CodeToHTML.C_COMMENTS and startIdx + 1 < len(code) and code[startIdx + 1] == "/")):
                endIdx = tokenEnd(code, endIdx, lambda x: x != "\n" and x != "\r")
                code, startIdx = subst(code, CodeToHTML.CLASS_COMMENT, startIdx, endIdx)
            elif (c == "/" and self.commentType == CodeToHTML.C_COMMENTS and startIdx + 1 < len(code) and code[startIdx + 1] == "*"):
                endIdx = code.find("*/", startIdx)
                if endIdx >= 0:
                    endIdx += 2
                else:  # unterminated C-style comment
                    endIdx = len(code) - 1
                code, startIdx = subst(code, CodeToHTML.CLASS_COMMENT, startIdx, endIdx)
            elif c == '"':
                endIdx = tokenEnd(code, endIdx, lambda x: x != '"')
                code, startIdx = subst(code, CodeToHTML.CLASS_CONSTANT, startIdx, endIdx)
            elif c.isalpha() or c == "_":
                endIdx = tokenEnd(code, endIdx, lambda x: x.isalnum() or x == "_")
                token = code[startIdx:endIdx]
                if token in self.typeSet: # query type so it can override keyword
                    code, startIdx = subst(code, CodeToHTML.CLASS_TYPE, startIdx, endIdx)
                elif token in self.constantSet:
                    code, startIdx = subst(code, CodeToHTML.CLASS_CONSTANT, startIdx, endIdx)
                elif token in self.keywordSet:
                    code, startIdx = subst(code, CodeToHTML.CLASS_KEYWORD, startIdx, endIdx)
                elif token in self.functionSet:
                    code, startIdx = subst(code, CodeToHTML.CLASS_FUNCTION_CALL, startIdx, endIdx)
                elif endIdx < len(code) and code[endIdx] == "(":
                    code, startIdx = subst(code, CodeToHTML.CLASS_FUNCTION_CALL, startIdx, endIdx)
                else:
                    startIdx = endIdx
            elif c.isdigit():
                endIdx = tokenEnd(code, endIdx, lambda x: x.isdigit() or x in [".", "e", "E"])
                if startIdx > 0 and code[startIdx - 1] in ["-", "+"]:
                    startIdx -= 1
                code, startIdx = subst(code, CodeToHTML.CLASS_CONSTANT, startIdx, endIdx)

            startIdx += 1
        
        return code

    def parseXML(self, code):
        def maketype(code, clazz, startIdx, endIdx):
            text = code[startIdx:endIdx]
            text = text.replace("<", "&lt")
            return '<span class="' + clazz + '">' + text + '</span>'

        html = ""
        startIdx = 0
        endIdx = 0
        while startIdx < len(code) and (code[startIdx] == '\n' or code[startIdx] == '\r'):
            startIdx += 1

        while startIdx < len(code):
            if code[startIdx] == "<":
                endIdx = startIdx + 1
                while endIdx < len(code) and code[endIdx] != ">":
                    endIdx += 1
                endIdx += 1
                if code[startIdx:endIdx].startswith("<!--"):
                    endIdx = code.find("-->", startIdx) + 3
                    if endIdx == -1:
                        endIdx = len(code)
                    html = html + maketype(code, CodeToHTML.CLASS_COMMENT, startIdx, endIdx)
                else:
                    firstSpaceIdx = startIdx
                    while firstSpaceIdx < endIdx and not code[firstSpaceIdx].isspace():
                        firstSpaceIdx += 1
                    html = html + maketype(code, CodeToHTML.CLASS_KEYWORD, startIdx, firstSpaceIdx)
                    text = code[firstSpaceIdx:endIdx].replace("<", "&lt;")
                    html = html + text

                startIdx = endIdx
            else:
                html = html + code[startIdx]
                startIdx += 1

        return html

#-------------------------------------------------------------------------------
main()
